
package uk.co.badgersinfoil.jactionscript;

import java.util.Iterator;


/**
 * An ActionList declares a list of actionscript bytecodes, in execution
 * order.  The list doesn't store {@link Action} objects directly, they are
 * referred to indirectly via an {@link ActionHandle} object.  This extra level
 * of indirection allows the list to manage the targets of branch operations
 * in the face of insertions or deletions (branch operations keep a reference
 * to the ActionHandle of their target, rather then the target Action itself, or
 * an integer offset into the list).  ActionHandle objects form a doubly-linked
 * list that should be managed though calls to the ActionList that owns them.
 *
 * <p>Insertion of actions into the middle of the list isn't implemented
 * yet.  Sorry.</p>
 */
public class ActionList {
	private ActionHandle first = null;
	private ActionHandle last = null;
	private ActionHandle next = new ActionHandle();
	private int length = 0;

	public ActionHandle append(Action op) {
		next.setAction(op);
		ActionHandle handle = append(next);
		next = new ActionHandle();
		return handle;
	}

	private ActionHandle append(ActionHandle handle) {
		handle.setPrev(last);
		handle.setNext(null);
		if (last == null) {
			first = handle;
		} else {
			last.setNext(handle);
		}
		last = handle;
		length++;
		return handle;
	}

	/**
	 * Returns a reference to the object that will be returned by the next
	 * call to {@link #append(Action)}.
	 */
	public ActionHandle peekNext() {
		return next;
	}

	/**
	 * Set the ActionHandle for the next Action appended.  This can be used when
	 * you want to create a forward branch like so:<pre>
	 *   ActionHandle target = new ActionHandle();
	 *   list.append(new IfJumpAction(target));
	 *   <i>..more appends..</i>
	 *   list.pokeNext(target);
	 *   list.append(<i>..the target Action..</i>);
	 * </pre>
	 */
	public void pokeNext(ActionHandle handle) {
		next = handle;
	}

	public void accept(Visitor v) throws JASVisitorException {
		ActionHandle current = first;
		while (current != null) {
			current.accept(v);
			Action op = current.getAction();
			op.accept(v);
			if (op instanceof StartFunctionAction) {
				ActionList body = ((StartFunctionAction)op).getBody();
				body.accept(v);
				v.visitEnd(body);
			}
			current = current.getNext();
		}
		// generate any final label,
		next.accept(v);
	}

	public int length() {
		return length;
	}

	/**
	 * Returns the handle on the first Action in the list
	 */
	public ActionHandle first() {
		return first;
	}

	/**
	 * Returns the handle on the last Action in the list
	 */
	public ActionHandle last() {
		return last;
	}

	/**
	 * @throws IllegalArgumentException if pos if less than 0, or beond
	 *         the end of the list
	 */
	public ActionHandle get(int pos) {
		if (pos<0 || pos>=length) {
			throw new IllegalArgumentException();
		}
		// could optimize when pos > length/2,
		ActionHandle result = first;
		while (pos-- > 0) {
			result = result.getNext();
		}
		return result;
	}

	/**
	 * Removes the Action referenced by the given handle from this list.
	 * If this handle was the target of any {@link BranchAction} objects,
	 * those branches will be re-targeted at the next ActionHandle in the
	 * list (or the handle given by {@link #peekNext()} if you are removing
	 * the last item in the list).
	 *
	 * <p>Trying to remove a handle not in this list will result in
	 * undefined behaviour.</p>
	 */
	public void remove(ActionHandle handle) {
		ActionHandle lprev = handle.getPrev();
		ActionHandle lnext = handle.getNext();
		if (handle == first) {
			first = lnext;
		}
		if (handle == last) {
			last = lprev;
		}
		if (lnext != null) {
			lnext.setPrev(lprev);
		}
		if (lprev != null) {
			lprev.setNext(lnext);
		}
		length--;
		if (handle.hasTargeters()) {
			if (lnext == null) {
				lnext = next;
			}
			Iterator i = handle.targeters();
			while (i.hasNext()) {
				BranchAction branch = (BranchAction)i.next();
				branch.setTarget(lnext);
			}
		}
		handle.dispose();
	}
}
