
package uk.co.badgersinfoil.jactionscript.analysis;


/**
 * Data on which Actions operate, either arguments poped of the emulated stack,
 * or results pushed back onto it.
 *
 * <p>The value of an operand may not be known, as the analysis isn't able to
 * determine the the result of actions in many cases.  {@link #knownValue()}
 * should be checked before using the operand's value.</p>
 *
 * <p>Actionscript's 'undefined' value may be a source of confusion, because
 * 'undefined' is a concrete, knowm value, making it different from an
 * Operand whose value is 'unknown'.</p>
 */
public class Operand {
	public static final class ASType {
		private String desc;

		private ASType(String desc) {
			this.desc = desc;
		}

		public static final ASType BOOLEAN = new ASType("boolean");
		public static final ASType DOUBLE = new ASType("double");
		public static final ASType STRING = new ASType("string");
		public static final ASType OBJECT = new ASType("object");
		public static final ASType FLOAT = new ASType("float");
		public static final ASType INTEGER = new ASType("integer");
		public static final ASType NULL = new ASType("null");
		public static final ASType UNDEFINED = new ASType("undefined");
		public static final ASType FUNCTION = new ASType("function");
		public static final ASType MOVIECLIP = new ASType("movieclip");

		public String toString() { return "["+desc+"]"; }
	}

	private Object val = null;
	private ASType type = null;

	public boolean knownValue() {
		return val != null;
	}
	public boolean knownType() {
		return type != null;
	}
	public void setValue(Object val) {
		this.val = val;
	}
	public void setType(ASType type) {
		this.type = type;
	}
	public Object getValue() {
		return val;
	}
	public ASType getType() {
		return type;
	}
	public void unsetValue() {
		val = null;
	}
	public void unsetType() {
		type = null;
	}
	public String toString() {
		if (knownType()) {
			if (knownValue()) {
				return "[Operand "+getType()+" "+getValue()+"]";
			} else {
				return "[Operand "+getType()+" (unkonwn value)]";
			}
		} else {
			return "[Operand (unknown)]";
		}
	}

	public static Operand mkundefined() {
		Operand entry = new Operand();
		entry.setType(ASType.UNDEFINED);
		entry.setValue(ASType.UNDEFINED);
		return entry;
	}

	public static Operand mknull() {
		Operand entry = new Operand();
		entry.setType(ASType.NULL);
		return entry;
	}

	public static Operand mkunknown() {
		Operand entry = new Operand();
		return entry;
	}

	/**
	 * Build a stack entry which is the result of poping from an empty
	 * stack.  Currently a synonym for {@link #mkunknown()}.
	 */
	public static Operand mkunderrun() {
		return mkunknown();
	}

	public static Operand mkstring(String s) {
		Operand entry = new Operand();
		entry.setValue(s);
		entry.setType(ASType.STRING);
		return entry;
	}

	public static Operand mkinteger(int i) {
		return mkinteger(new Integer(i));
	}

	public static Operand mkinteger(Integer i) {
		Operand entry = new Operand();
		entry.setValue(i);
		entry.setType(ASType.INTEGER);
		return entry;
	}

	public static Operand mkfloat(float f) {
		return mkfloat(new Float(f));
	}

	public static Operand mkfloat(Float f) {
		Operand entry = new Operand();
		entry.setValue(f);
		entry.setType(ASType.FLOAT);
		return entry;
	}

	public static Operand mkboolean(boolean b) {
		return mkboolean(b ? Boolean.TRUE : Boolean.FALSE);
	}

	public static Operand mkboolean(Boolean b) {
		Operand entry = new Operand();
		entry.setValue(b);
		entry.setType(ASType.BOOLEAN);
		return entry;
	}

	public static Operand mkboolean() {
		Operand entry = new Operand();
		entry.setType(ASType.BOOLEAN);
		return entry;
	}

	public static Operand mkdouble(Double d) {
		Operand entry = new Operand();
		entry.setValue(d);
		entry.setType(ASType.DOUBLE);
		return entry;
	}
}
