
package uk.co.badgersinfoil.jactionscript.analysis;

import uk.co.badgersinfoil.jactionscript.*;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

// TODO: remove scraps of code that try to calculate results for a few actions
//       and place them in a StackListener implementation

/**
 * Builds a picture of the actionscript stack layout as each instruction is
 * executed, driving an implementation of {@link StackListener}.
 *
 * <p>For example, when a NOT action is encountered, we pop an Operand off
 * our emulated stack, create a new Operand to hold the result, and then
 * pass these values to {@link StackListener#visitNOT(Action,Operand,Operand)}.
 * The result-Operand's value isn't calculated, though the StackListener
 * implementation may fill this in.  When visitNot() returns, the
 * SimpleStackSimulator pushes the result Operand onto the stack.</p>
 *
 * <p>The only time that this class fills in the value for Operands is where
 * constant values are pushed onto the stack.  Without this, we wouldn't
 * be able to resonably emulate things like function-calls, where the number
 * of values to pop off the stack is determined by the (usually constant)
 * value on the top of the stack.</p>
 *
 * <p>Since no attempt is made to follow the branches in the flow of control
 * produced by {@link BranchAction} subclasses, encountering any Action that is
 * a branch, or that is the target of a branch invalidates all our guesses
 * about the stack layout.  Later attempts to read from this invalid part
 * of the stack will always produce Operands with unknown values.</p>
 */
public class SimpleStackSimulator implements Visitor {
	protected StackInfo stack = new StackInfo();
	protected LookupTable currentLookupTable = null;
	private StackListener listener;


	public SimpleStackSimulator(StackListener listener) {
		this.listener = listener;
	}

	/** convenience method; calls stack.push(entry) */
	protected void push(Operand entry) {
		stack.push(entry);
	}

	/** convenience method; calls stack.pop() */
	protected Operand pop() {
		return stack.pop();
	}

	protected void invalidateState() {
		stack.invalidate();
		// TODO: lookup table probably becomes invalid
		//currentLookupTable = null;
	}


	/**
	 * pops a list of arguments off the stack ant returns the Operand
	 * objects in a List.  The number of arguments is expected as the value
	 * on the top of the stack; after the argument count is poped, the
	 * number of values indicated are poped and added to the result List.
	 * If the arguemnt count is 0, this mathod will return null, rather
	 * than an empty List.
	 */
	protected List popArgs() {
		Operand argCount = pop();
		if (!argCount.knownValue()) {
			throw new RuntimeException("number of arguments unknown at analysis-time");
		}
		int count = ((Number)argCount.getValue()).intValue();
		List args = null;
		if (count > 0) {
			args = new ArrayList();
			for (int i=0 ; i<count ; i++) {
				Operand arg = pop();
				args.add(arg);
			}
		}
		return args;
	}

	public void visitHandle(ActionHandle handle)
		throws JASVisitorException
	{
		// TODO: need to invalidateState() is this is a jump target
	}

	public void visitEnd(ActionList list)
		throws JASVisitorException
	{
		
	}

	public void visitADD(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.DOUBLE);
		listener.visitADD(action, left, right, result);
		push(result);
	}

	public void visitAND(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = Operand.mkboolean();
		listener.visitAND(action, left, right, result);
		push(result);
	}

	public void visitASCII_TO_CHAR(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = Operand.mkunknown();
		listener.visitASCII_TO_CHAR(action, val, result);
		push(result);
	}

	public void visitASCII_TO_CHAR_MB(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = Operand.mkunknown();
		listener.visitASCII_TO_CHAR_MB(action, val, result);
		push(result);
	}

	public void visitBIT_AND(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitBIT_AND(action, left, right, result);
		push(result);
	}

	public void visitBIT_OR(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitBIT_OR(action, left, right, result);
		push(result);
	}

	public void visitBIT_XOR(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitBIT_XOR(action, left, right, result);
		push(result);
	}

	public void visitCALL(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitCALL_FUNCTION(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		List args = popArgs();
		Operand result = Operand.mkunknown();
		listener.visitCALL_FUNCTION(action, name, args, result);
		push(result);
	}

	public void visitCALL_METHOD(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		Operand obj = pop();
		List args = popArgs();
		Operand result = Operand.mkunknown();
		listener.visitCALL_METHOD(action, name, args, result);
		push(result);
	}

	public void visitCHAR_MB_TO_ASCII(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitCHAR_MB_TO_ASCII(action, val, result);
		push(result);
	}

	public void visitCHAR_TO_ASCII(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitCHAR_MB_TO_ASCII(action, val, result);
		push(result);
	}

	public void visitCLONE_SPRITE(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitCONCAT(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.STRING);
		listener.visitCONCAT(action, left, right, result);
		push(result);
	}

	public void visitCONVERT_TO_NUMBER(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitCONVERT_TO_NUMBER(action, val, result);
		push(result);
	}

	public void visitCONVERT_TO_STRING(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.STRING);
		listener.visitCONVERT_TO_STRING(action, val, result);
		push(result);
	}

	public void visitDECREMENT(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitDECREMENT(action, val, result);
		push(result);
	}

	public void visitDEFINE_LOCAL(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		listener.visitDEFINE_LOCAL(action, name);
	}

	public void visitDEFINE_LOCAL_VALUE(Action action)
		throws JASVisitorException
	{
		Operand value = pop();
		Operand name = pop();
		listener.visitDEFINE_LOCAL_VALUE(action, name, value);
	}

	public void visitDELETE_PROPERTY(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		Operand obj = pop();
		Operand result = Operand.mkunknown();
		listener.visitDELETE_PROPERTY(action, name, obj, result);
		push(result);
	}

	public void visitDELETE_THREAD_VARS(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		listener.visitDELETE_THREAD_VARS(action, name);
	}

	public void visitDIVIDE(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = Operand.mkunknown();
		listener.visitDIVIDE(action, left, right, result);
		push(result);
	}

	public void visitDUPLICATE(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result1 = val/*.clone()*/;
		Operand result2 = val/*.clone()*/;
		listener.visitDUPLICATE(action, val, result1, result2);
		push(result1);
		push(result2);
	}

	public void visitEND_DRAG(Action action)
		throws JASVisitorException
	{
		listener.visitEND_DRAG(action);
	}

	public void visitENUMERATE(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		push(Operand.mknull()); // null indicates end of slot list

		List slots = new ArrayList();
		listener.visitENUMERATE(action, name, slots);
		Iterator i = slots.iterator();
		while (i.hasNext()) {
			push((Operand)i.next());
		}

		// The layout of the stack is now probably unknown.  Maybe
		// some slots will have been pushed, maybe not.  How should
		// we denote this?
	}

	public void visitENUMERATE_OBJECT(Action action)
		throws JASVisitorException
	{
		Operand obj = pop();
		push(Operand.mknull()); // null indicates end of slot list

		List slots = new ArrayList();
		listener.visitENUMERATE_OBJECT(action, obj, slots);
		Iterator i = slots.iterator();
		while (i.hasNext()) {
			push((Operand)i.next());
		}

		// The layout of the stack is now probably unknown.  Maybe
		// some slots will have been pushed, maybe not.  How should
		// we denote this?
	}

	public void visitEQUALS(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = Operand.mkboolean();
		listener.visitEQUALS(action, left, right, result);
		push(result);
	}

	public void visitGET_MEMBER(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		Operand obj = pop();
		Operand result = Operand.mkunknown();
		listener.visitGET_MEMBER(action, name, obj, result);
		push(result);
	}

	public void visitGET_PROPERTY(Action action)
		throws JASVisitorException
	{
		Operand index = pop();
		Operand target = pop();
		Operand result = Operand.mkunknown();
		listener.visitGET_MEMBER(action, index, target, result);
		push(result);
	}

	public void visitGET_TARGET_PATH(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitGET_TIME(Action action)
		throws JASVisitorException
	{
		Operand time = new Operand();
		time.setType(Operand.ASType.INTEGER);
		listener.visitGET_TIME(action, time);
		push(time);
	}

	public void visitGET_URL2(GetURL2Action action)
		throws JASVisitorException
	{
		Operand target = pop();
		Operand url = pop();
		listener.visitGET_URL2(action, target, url);
		// TODO: if target is a sprite, rather than a browser
		//       window, we could clobber the sprites variables
		//       (not that we maintain the sprites variables in any
		//       case)
	}

	public void visitGET_URL(GetURLAction action)
		throws JASVisitorException
	{
		listener.visitGET_URL(action);
		// no effect on stack
	}

	public void visitGET_VARIABLE(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		Operand result = Operand.mkunknown();
		listener.visitGET_VARIABLE(action, name, result);
		push(result);
	}

	public void visitGOTO_FRAME2(GotoFrame2Action action)
		throws JASVisitorException
	{
		Operand frame = pop();
		listener.visitGOTO_FRAME2(action, frame);
	}

	public void visitGOTO_FRAME(GotoFrameAction action)
		throws JASVisitorException
	{
		listener.visitGOTO_FRAME(action);
		// no effect on stack
	}

	public void visitGOTO_LABEL(GotoLabelAction action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitGREATER_THAN(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkunknown();
		listener.visitGREATER_THAN(action, left, right, result);
		push(result);
	}

	public void visitIF_JUMP(IfJumpAction action)
		throws JASVisitorException
	{
		Operand condition = pop();
		listener.visitIF_JUMP(action, condition);
	}

	public void visitINCREMENT(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitINCREMENT(action, val, result);
		push(result);
	}

	public void visitINIT_ARRAY(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitINIT_OBJECT(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitINSTANCE_OF(Action action)
		throws JASVisitorException
	{
		Operand ctor = pop();
		Operand obj = pop();
		Operand result = Operand.mkunknown();
		listener.visitINSTANCE_OF(action, ctor, obj, result);
		push(result);
	}

	public void visitJUMP(JumpAction action)
		throws JASVisitorException
	{
		invalidateState();
		listener.visitJUMP(action);
	}

	public void visitLESS_THAN(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkunknown();
		listener.visitINSTANCE_OF(action, left, right, result);
		push(result);
	}

	// TODO: no LESS_THAN2 ?

	public void visitLOOKUP(LookupAction action)
		throws JASVisitorException
	{
		Operand result;
		if (currentLookupTable == null) {
			result = Operand.mkunknown();
		} else {
			String val = currentLookupTable.lookup(action.getIndex());
			result = Operand.mkstring(val);
		}
		listener.visitLOOKUP(action, result);
		push(result);
	}

	public void visitLOOKUP_TABLE(LookupTableAction action)
		throws JASVisitorException
	{
		currentLookupTable = action.getLookupTable();
		listener.visitLOOKUP_TABLE(action);
		// no effect on stack
	}

	public void visitMODULO(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.INTEGER);
		listener.visitMODULO(action, left, right, result);
		push(result);
	}

	public void visitMULTIPLY(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.DOUBLE);
		listener.visitMULTIPLY(action, left, right, result);
		push(result);
	}

	public void visitNEW_METHOD(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitNEW_OBJECT(Action action)
		throws JASVisitorException
	{
		Operand name = pop();
		List args = popArgs();
		Operand result = new Operand();
		result.setType(Operand.ASType.OBJECT);
		// TODO: anthing we can do with 'name'?
		//       e.g. set type of object?
		listener.visitNEW_OBJECT(action, name, args, result);
		push(result);
	}

	public void visitNEXT_FRAME(Action action)
		throws JASVisitorException
	{
		// no effect on stack
		listener.visitNEXT_FRAME(action);
	}

	public void visitNOT(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = Operand.mkboolean();
		listener.visitNOT(action, val, result);
		push(result);
	}

	public void visitOR(Action action)
		throws JASVisitorException
	{
		Operand left = pop();
		Operand right = pop();
		Operand result = Operand.mkboolean();
		listener.visitOR(action, left, right, result);
		push(result);
	}

	public void visitPLAY(Action action)
		throws JASVisitorException
	{
		listener.visitPLAY(action);
		// no effect on stack
	}

	public void visitPOP(Action action)
		throws JASVisitorException
	{
		pop();
		listener.visitPOP(action);
	}

	public void visitPREV_FRAME(Action action)
		throws JASVisitorException
	{
		listener.visitPREV_FRAME(action);
		// no effect on stack
	}

	public void visitPUSH(PushAction action)
		throws JASVisitorException
	{
		List list = new ArrayList();
		Iterator i = action.valueIterator();
		while (i.hasNext()) {
			Object v = i.next();
			if (v instanceof Boolean) {
				list.add(Operand.mkboolean((Boolean)v));
			} else if (v instanceof Double) {
				list.add(Operand.mkdouble((Double)v));
			} else if (v instanceof Float) {
				list.add(Operand.mkfloat((Float)v));
			} else if (v instanceof Integer) {
				list.add(Operand.mkinteger((Integer)v));
			} else if (v instanceof String) {
				list.add(Operand.mkstring((String)v));
			} else if (v == null) {
				list.add(Operand.mknull());
			} else if (v instanceof Register) {
				// TODO: better register handling
				list.add(Operand.mkunknown());
			} else if (v == PushAction.UNDEFINED) {
				list.add(Operand.mkundefined());
			} else {
				throw new Error("unexpected push value: "+v);
			}
		}
		listener.visitPUSH(action, list);
		i = list.iterator();
		while (i.hasNext()) {
			push((Operand)i.next());
		}
	}

	public void visitRANDOM_NUMBER(Action action)
		throws JASVisitorException
	{
		Operand max = pop();
		Operand result = new Operand();
		// TODO: not sure if random numbers are float, or something
		//       determined by the max value,
		result.setType(Operand.ASType.FLOAT);
		listener.visitRANDOM_NUMBER(action, max, result);
		push(result);
	}

	public void visitREMOVE_SPRITE(Action action)
		throws JASVisitorException
	{
		Operand target = pop();
		listener.visitREMOVE_SPRITE(action, target);
	}

	public void visitRETURN_VALUE(Action action)
		throws JASVisitorException
	{
		Operand returnVal = pop();
		listener.visitREMOVE_SPRITE(action, returnVal);
	}

	public void visitSET_MEMBER(Action action)
		throws JASVisitorException
	{
		Operand value = pop();
		Operand name = pop();
		Operand object = pop();
		listener.visitSET_MEMBER(action, value, name, object);
	}

	public void visitSET_PROPERTY(Action action)
		throws JASVisitorException
	{
		Operand value = pop();
		Operand index = pop();
		Operand target = pop();
		listener.visitSET_PROPERTY(action, value, index, target);
	}

	public void visitSET_TARGET_2(Action action)
		throws JASVisitorException
	{
		Operand target = pop();
		listener.visitSET_TARGET_2(action, target);
	}

	public void visitSET_TARGET(SetTargetAction action)
		throws JASVisitorException
	{
		listener.visitSET_TARGET(action);
		// no effect on stack
	}

	public void visitSET_VARIABLE(Action action)
		throws JASVisitorException
	{
		Operand value = pop();
		Operand name = pop();
		listener.visitSET_VARIABLE(action, name, value);
	}

	public void visitSHIFT_LEFT(Action action)
		throws JASVisitorException
	{
		Operand count = pop();
		Operand value = pop();
		Operand result = Operand.mkunknown();
		listener.visitSHIFT_LEFT(action, count, value, result);
		push(result);
	}

	public void visitSHIFT_RIGHT(Action action)
		throws JASVisitorException
	{
		Operand count = pop();
		Operand value = pop();
		Operand result = Operand.mkunknown();
		listener.visitSHIFT_RIGHT(action, count, value, result);
		push(result);
	}

	public void visitSHIFT_RIGHT_UNSIGNED(Action action)
		throws JASVisitorException
	{
		Operand count = pop();
		Operand value = pop();
		Operand result = Operand.mkunknown();
		listener.visitSHIFT_RIGHT_UNSIGNED(action, count, value, result);
		push(result);
	}

	public void visitSTART_DRAG(Action action)
		throws JASVisitorException
	{
		Operand target = pop();
		Operand lookCentre = pop();
		Operand constrain = pop();
		if (!constrain.knownValue()) {
			// TODO: better to just invalidate the stack?
			throw new RuntimeException("ASSERT: value of 'constrain' not known at analyse-time, can't continue.  Sorry this isn't handled better");
		}
		Object val = constrain.getValue();
		StackListener.DragBox constraintBox = null;
		if (val instanceof Integer && ((Integer)val).intValue()!=0
		  ||val instanceof Boolean && ((Boolean)val).booleanValue())
		{
			constraintBox = new StackListener.DragBox(pop(), pop(),
			                                          pop(), pop());
		}
		listener.visitSTART_DRAG(action, target, lookCentre, constraintBox);
	}

	public void visitSTART_FUNCTION(StartFunctionAction action)
		throws JASVisitorException
	{
		Operand fn;
		if (action.getName().length() == 0) {
			// anonymous function; reference to the function is
			// pushed onto the stack
			fn = new Operand();
			fn.setType(Operand.ASType.FUNCTION);
			push(fn);
		} else {
			fn = null;
			// TODO: somehow note that this name is now defined
		}
		listener.visitSTART_FUNCTION(action, fn);
	}

	public void visitSTART_WITH(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitSTOP(Action action)
		throws JASVisitorException
	{
		listener.visitSTOP(action);
		// no effect on stack
	}

	public void visitSTOP_SOUNDS(Action action)
		throws JASVisitorException
	{
		listener.visitSTOP_SOUNDS(action);
		// no effect on stack
	}

	public void visitSTORE_IN_REGISTER(StoreInRegisterAction action)
		throws JASVisitorException
	{
		Operand val = stack.peek();
		listener.visitSTORE_IN_REGISTER(action, val);
		// stack unchanged, we don't remove the peek'ed value
	}

	public void visitSTRICT_EQUALS(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkboolean();
		listener.visitSTRICT_EQUALS(action, left, right, result);
		push(result);
	}

	public void visitSTRING_EQUALS(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result;
		if (!left.knownValue() || !right.knownValue()) {
			result = Operand.mkunknown();
		} else if (left.getType() != Operand.ASType.STRING
		        || right.getType() != Operand.ASType.STRING)
		{
			// TODO: what does flash do if values aren't [String]?
			result = Operand.mkundefined();
		} else {
			result = Operand.mkboolean(left.getValue()
			                           .equals(right.getValue()));
		}
		listener.visitSTRING_EQUALS(action, left, right, result);
		push(result);
	}

	public void visitSTRING_GREATER_THAN(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkboolean();
		listener.visitSTRING_GREATER_THAN(action, left, right, result);
		push(result);
	}

	public void visitSTRING_LENGTH(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = Operand.mkunknown();
		listener.visitSTRING_LENGTH(action, val, result);
		push(result);
	}

	public void visitSTRING_LENGTH_MB(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = Operand.mkunknown();
		listener.visitSTRING_LENGTH_MB(action, val, result);
		push(result);
	}

	public void visitSTRING_LESS_THAN(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkunknown();
		listener.visitSTRING_LESS_THAN(action, left, left, result);
		push(result);
	}

	public void visitSUBSTRING(Action action)
		throws JASVisitorException
	{
		Operand count = pop();
		Operand index = pop();
		Operand val = pop();
		Operand result;
		if (!count.knownValue() || !index.knownValue()
		  || !val.knownValue())
		{
			result = Operand.mkunknown();
		} else if (count.getType() != Operand.ASType.INTEGER
		          ||index.getType() != Operand.ASType.INTEGER) {
			// TODO: is float acceptable too?
			result = Operand.mkstring("");
		} else if (val.getType() == Operand.ASType.UNDEFINED) {
			result = Operand.mkundefined();
		} else {
			// TODO: implement substring
			result = new Operand();
			result.setType(Operand.ASType.STRING);
		}
		listener.visitSUBSTRING(action, count, index, val, result);
		push(result);
	}

	public void visitSUBSTRING_MB(Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitSUBTRACT(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result;
		if (!left.knownValue() || !right.knownValue()) {
			result = new Operand();
			result.setType(Operand.ASType.FLOAT);
		} else {
			float l;
			if (left.getType() == Operand.ASType.INTEGER
			   || left.getType() == Operand.ASType.FLOAT)
			{
				l = ((Number)left.getValue()).floatValue();
			} else {
				l = 0;
			}
			float r;
			if (right.getType() == Operand.ASType.INTEGER
			   || right.getType() == Operand.ASType.FLOAT)
			{
				r = ((Number)right.getValue()).floatValue();
			} else {
				r = 0;
			}
			result = Operand.mkfloat(l - r);
		}
		listener.visitSUBTRACT(action, left, right, result);
		push(result);

	}

	public void visitSWAP(Action action)
		throws JASVisitorException
	{
		Operand first = pop();
		Operand second = pop();
		listener.visitSWAP(action, first, second);
		push(first);
		push(second);
	}

	public void visitTOGGLE_QUALITY(Action action)
		throws JASVisitorException
	{
		listener.visitTOGGLE_QUALITY(action);
		// no effect on stack
	}

	public void visitTO_INTEGER(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result;
		if (!val.knownValue()) {
			result = Operand.mkunknown();
		} else if (val.getType() == Operand.ASType.INTEGER) {
			result = val;  // TODO: duplicate instead?
		} else if (val.getType() == Operand.ASType.FLOAT) {
			int i = ((Float)val.getValue()).intValue();
			result = Operand.mkinteger(i);
		} else if (val.getType() == Operand.ASType.STRING) {
			Integer i;
			try {
				i = Integer.valueOf((String)val.getValue());
			} catch (NumberFormatException e) {
				i = new Integer(0);
			}
			result = Operand.mkinteger(i);
		} else {
			throw new Error("unimplemented: conversion to integer from "+val.getType());
		}
		listener.visitTO_INTEGER(action, val, result);
		push(result);
	}

	public void visitTRACE(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		listener.visitTRACE(action, val);
	}

	public void visitTYPED_ADD(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkunknown();
		listener.visitTYPED_ADD(action, left, right, result);
		push(result);
	}

	public void visitTYPED_EQUALS(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkboolean();
		listener.visitTYPED_EQUALS(action, left, right, result);
		push(result);
	}

	public void visitTYPED_LESS_THAN(Action action)
		throws JASVisitorException
	{
		Operand right = pop();
		Operand left = pop();
		Operand result = Operand.mkboolean();
		listener.visitTYPED_LESS_THAN(action, left, right, result);
		push(result);
	}

	public void visitTYPE_OF(Action action)
		throws JASVisitorException
	{
		Operand val = pop();
		Operand result = new Operand();
		result.setType(Operand.ASType.STRING);
		// TODO: implement some string values
		listener.visitTYPE_OF(action, val, result);
		push(result);
	}

	public void visitUNKNOWN(UnknownAction action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitWAIT_FOR_FRAME(WaitForFrameAction action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

	public void visitWAIT_FOR_FRAME_2(WaitForFrame2Action action)
		throws JASVisitorException
	{
		throw new RuntimeException("No implemented");
	}

}
